@@ -3,20 +3,24 @@
  *
  * All rights reserved.
  *
- * This file is part of ObjFW. It may be distributed under the terms of the
- * Q Public License 1.0, which can be found in the file LICENSE.QPL included in
- * the packaging of this file.
+ * This program is free software: you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License version 3.0 only,
+ * as published by the Free Software Foundation.
  *
- * Alternatively, it may be distributed under the terms of the GNU General
- * Public License, either version 2 or 3, which can be found in the file
- * LICENSE.GPLv2 or LICENSE.GPLv3 respectively included in the packaging of this
- * file.
+ * This program is distributed in the hope that it will be useful, but WITHOUT
+ * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+ * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
+ * version 3.0 for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public License
+ * version 3.0 along with this program. If not, see
+ * <https://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
 
 #ifndef _LARGEFILE64_SOURCE
-# define _LARGEFILE64_SOURCE
+# define _LARGEFILE64_SOURCE 1
 #endif
 
 #include <errno.h>
@@ -35,6 +39,12 @@
 #if defined(OF_LINUX) || defined(OF_MACOS)
 # include <sys/xattr.h>
 #endif
+#if defined(OF_FREEBSD) || defined(OF_NETBSD)
+# include <sys/extattr.h>
+#endif
+#ifdef OF_HAIKU
+# include <kernel/fs_attr.h>
+#endif
 #ifdef OF_WINDOWS
 # include <utime.h>
 #endif
@@ -147,6 +157,11 @@ static WINAPI BOOLEAN (*createHardLinkWFuncPtr)(LPCWSTR, LPCWSTR,
     LPSECURITY_ATTRIBUTES);
 #endif
 
+#ifdef OF_FREEBSD
+static const char *namespaces[] = EXTATTR_NAMESPACE_NAMES;
+static int numNamespaces = sizeof(namespaces) / sizeof(*namespaces);
+#endif
+
 #ifdef OF_WINDOWS
 static OFTimeInterval
 filetimeToTimeInterval(const FILETIME *filetime)
@@ -165,6 +180,8 @@ lastError(void)
 		return ENOENT;
 	case ERROR_ACCESS_DENIED:
 		return EACCES;
+	case ERROR_PRIVILEGE_NOT_HELD:
+		return EPERM;
 	case ERROR_DIRECTORY:
 		return ENOTDIR;
 	case ERROR_NOT_READY:
@@ -372,6 +389,38 @@ lstatWrapper(OFString *path, Stat *buffer)
 #endif
 }
 
+#if defined(OF_FREEBSD) || defined(OF_NETBSD)
+static void
+parseAttributeName(OFString **name, int *namespace)
+{
+	size_t pos = [*name rangeOfString: @"."].location;
+	OFString *namespaceName;
+	const char *cNamespace;
+
+	if (pos == OFNotFound)
+		@throw [OFInvalidArgumentException exception];
+
+	namespaceName = [*name substringToIndex: pos];
+	cNamespace = [namespaceName cStringWithEncoding: [OFLocale encoding]];
+
+	*name = [*name substringFromIndex: pos + 1];
+
+# if defined(OF_FREEBSD)
+	for (int i = 0; i < numNamespaces; i++) {
+		if (strcmp(namespaces[i], cNamespace) == 0) {
+			*namespace = i;
+			return;
+		}
+	}
+
+	@throw [OFInvalidArgumentException exception];
+# elif defined(OF_NETBSD)
+	if (extattr_string_to_namespace(cNamespace, namespace) == -1)
+		@throw [OFInvalidArgumentException exception];
+# endif
+}
+#endif
+
 static void
 setTypeAttribute(OFMutableFileAttributes attributes, Stat *s)
 {
@@ -585,22 +634,27 @@ setExtendedAttributes(OFMutableFileAttributes attributes, OFIRI *IRI)
 	OFString *path = IRI.fileSystemRepresentation;
 	OFStringEncoding encoding = [OFLocale encoding];
 	const char *cPath = [path cStringWithEncoding: encoding];
-# if defined(OF_LINUX)
-	ssize_t size = llistxattr(cPath, NULL, 0);
-# elif defined(OF_MACOS)
-	ssize_t size = listxattr(cPath, NULL, 0, XATTR_NOFOLLOW);
-# endif
-	char *list = OFAllocMemory(1, size);
 	OFMutableArray *names = nil;
+# if defined(OF_LINUX) || defined(OF_MACOS)
+#  if defined(OF_LINUX)
+	ssize_t size = llistxattr(cPath, NULL, 0);
+#  elif defined(OF_MACOS)
+	ssize_t size = listxattr(cPath, NULL, 0, XATTR_NOFOLLOW);
+#  endif
+	char *list;
 
+	if (size < 0)
+		return;
+
+	list = OFAllocMemory(1, size);
 	@try {
 		char *name;
 
-# if defined(OF_LINUX)
+#  if defined(OF_LINUX)
 		if ((size = llistxattr(cPath, list, size)) < 0)
-# elif defined(OF_MACOS)
+#  elif defined(OF_MACOS)
 		if ((size = listxattr(cPath, list, size, XATTR_NOFOLLOW)) < 0)
-# endif
+#  endif
 			return;
 
 		names = [OFMutableArray array];
@@ -619,7 +673,90 @@ setExtendedAttributes(OFMutableFileAttributes attributes, OFIRI *IRI)
 	} @finally {
 		OFFreeMemory(list);
 	}
+# elif defined(OF_FREEBSD) || defined(OF_NETBSD)
+	names = [OFMutableArray array];
 
+#  if defined(OF_FREEBSD)
+	for (int i = 0; i < numNamespaces; i++) {
+		int namespace = i;
+		const char *cNamespace = namespaces[i];
+#  elif defined(OF_NETBSD)
+	for (size_t i = 0; extattr_namespaces[i] != 0; i++) {
+		int namespace = extattr_namespaces[i];
+		char *cNamespace;
+#  endif
+		ssize_t size;
+		char *list;
+
+		if ((size = extattr_list_link(cPath, namespace, NULL, 0)) < 0)
+			continue;
+
+		list = OFAllocMemory(1, size);
+		@try {
+			OFString *namespaceName;
+			char *iter;
+
+			if ((size = extattr_list_link(cPath, namespace,
+			    list, size)) < 0)
+				continue;
+
+#  ifdef OF_NETBSD
+			if (extattr_namespace_to_string(namespace,
+			    &cNamespace) == -1)
+				continue;
+#  endif
+
+			namespaceName = [OFString stringWithCString: cNamespace
+							   encoding: encoding];
+			iter = list;
+
+			while (size > 0) {
+				ssize_t length = *(unsigned char *)iter;
+				OFString *name;
+
+				iter++;
+				size--;
+
+				if (length > size)
+					@throw [OFOutOfRangeException
+					    exception];
+
+				name = [OFString stringWithCString: iter
+							  encoding: encoding
+							    length: length];
+				name = [OFString stringWithFormat:
+				    @"%@.%@",  namespaceName, name];
+
+				[names addObject: name];
+
+				iter += length;
+				size -= length;
+			}
+		} @finally {
+			OFFreeMemory(list);
+		}
+	}
+# elif defined(OF_HAIKU)
+	DIR *dir = fs_open_attr_dir(cPath);
+
+	if (dir == NULL)
+		return;
+
+	@try {
+		struct dirent *dirent;
+
+		names = [OFMutableArray array];
+
+		while ((dirent = fs_read_attr_dir(dir)) != NULL)
+			[names addObject:
+			    [OFString stringWithCString: dirent->d_name
+					       encoding: encoding]];
+	} @finally {
+		fs_close_attr_dir(dir);
+	}
+# endif
+
+	[names makeImmutable];
 	[attributes setObject: names forKey: OFFileExtendedAttributesNames];
 }
 #endif
@@ -1585,62 +1722,157 @@ setExtendedAttributes(OFMutableFileAttributes attributes, OFIRI *IRI)
 }
 
 #ifdef OF_FILE_MANAGER_SUPPORTS_EXTENDED_ATTRIBUTES
-- (OFData *)extendedAttributeDataForName: (OFString *)name
-			     ofItemAtIRI: (OFIRI *)IRI
-{
-	void *pool = objc_autoreleasePoolPush();
-	OFString *path = IRI.fileSystemRepresentation;
-	OFStringEncoding encoding = [OFLocale encoding];
-	const char *cPath = [path cStringWithEncoding: encoding];
-	const char *cName = [name cStringWithEncoding: encoding];
-# if defined(OF_LINUX)
-	ssize_t size = lgetxattr(cPath, cName, NULL, 0);
-# elif defined(OF_MACOS)
-	ssize_t size = getxattr(cPath, cName, NULL, 0, 0, XATTR_NOFOLLOW);
-# endif
-	void *value = OFAllocMemory(1, size);
-	OFData *data;
-
-	@try {
-# if defined(OF_LINUX)
-		if ((size = lgetxattr(cPath, cName, value, size)) < 0)
-# elif defined(OF_MACOS)
-		if ((size = getxattr(cPath, cName, value, size, 0,
-		    XATTR_NOFOLLOW)) < 0)
-# endif
-			@throw [OFGetItemAttributesFailedException
-			    exceptionWithIRI: IRI
-				       errNo: errno];
-
-		data = [OFData dataWithItems: value count: size];
-	} @finally {
-		OFFreeMemory(value);
-	}
-
-	[data retain];
-
-	objc_autoreleasePoolPop(pool);
-
-	return [data autorelease];
-}
-
-- (void)setExtendedAttributeData: (OFData *)data
+- (void)getExtendedAttributeData: (OFData **)data
+			 andType: (id *)type
 			 forName: (OFString *)name
 		     ofItemAtIRI: (OFIRI *)IRI
 {
 	void *pool = objc_autoreleasePoolPush();
 	OFString *path = IRI.fileSystemRepresentation;
 	OFStringEncoding encoding = [OFLocale encoding];
+	const char *cPath = [path cStringWithEncoding: encoding];
+	void *value = NULL;
+# if defined(OF_LINUX) || defined(OF_MACOS)
+	const char *cName = [name cStringWithEncoding: encoding];
+#  if defined(OF_LINUX)
+	ssize_t size = lgetxattr(cPath, cName, NULL, 0);
+#  elif defined(OF_MACOS)
+	ssize_t size = getxattr(cPath, cName, NULL, 0, 0, XATTR_NOFOLLOW);
+#  endif
 
-# if defined(OF_LINUX)
-	if (lsetxattr([path cStringWithEncoding: encoding],
-	    [name cStringWithEncoding: encoding], data.items,
-	    data.count * data.itemSize, 0) != 0) {
-# elif defined(OF_MACOS)
-	if (setxattr([path cStringWithEncoding: encoding],
-	    [name cStringWithEncoding: encoding], data.items,
-	    data.count * data.itemSize, 0, XATTR_NOFOLLOW) != 0) {
+	if (size < 0)
+		@throw [OFGetItemAttributesFailedException
+		    exceptionWithIRI: IRI
+			       errNo: errno];
+
+	value = OFAllocMemory(1, size);
+	@try {
+#  if defined(OF_LINUX)
+		if ((size = lgetxattr(cPath, cName, value, size)) < 0)
+#  elif defined(OF_MACOS)
+		if ((size = getxattr(cPath, cName, value, size, 0,
+		    XATTR_NOFOLLOW)) < 0)
+#  endif
+			@throw [OFGetItemAttributesFailedException
+			    exceptionWithIRI: IRI
+				       errNo: errno];
+
+		*data = [OFData dataWithItemsNoCopy: value
+					      count: size
+				       freeWhenDone: true];
+		value = NULL;
+	} @finally {
+		OFFreeMemory(value);
+	}
+
+	if (type != NULL)
+		*type = nil;
+# elif defined(OF_FREEBSD) || defined(OF_NETBSD)
+	int namespace;
+	const char *cName;
+	ssize_t size;
+
+	parseAttributeName(&name, &namespace);
+	cName = [name cStringWithEncoding: encoding];
+
+	if ((size = extattr_get_link(cPath, namespace, cName, NULL, 0)) < 0)
+		@throw [OFGetItemAttributesFailedException
+		    exceptionWithIRI: IRI
+			       errNo: errno];
+
+	value = OFAllocMemory(1, size);
+	@try {
+		if ((size = extattr_get_link(cPath, namespace, cName,
+		    value, size)) < 0)
+			@throw [OFGetItemAttributesFailedException
+			    exceptionWithIRI: IRI
+				       errNo: errno];
+
+		*data = [OFData dataWithItemsNoCopy: value
+					      count: size
+				       freeWhenDone: true];
+		value = NULL;
+	} @finally {
+		OFFreeMemory(value);
+	}
+
+	if (type != NULL)
+		*type = nil;
+# elif defined(OF_HAIKU)
+	const char *cName = [name cStringWithEncoding: encoding];
+	int fd = open(cPath, O_RDONLY);
+	struct attr_info info;
+
+	if (fd == -1)
+		@throw [OFGetItemAttributesFailedException
+		    exceptionWithIRI: IRI
+			       errNo: errno];
+
+	@try {
+		if (fs_stat_attr(fd, cName, &info) != 0)
+			@throw [OFGetItemAttributesFailedException
+			    exceptionWithIRI: IRI
+				       errNo: errno];
+
+		if (info.size < 0 || info.size > SSIZE_MAX)
+			@throw [OFOutOfRangeException exception];
+
+		value = OFAllocMemory(1, (size_t)info.size);
+
+		errno = 0;
+		if (fs_read_attr(fd, cName, B_ANY_TYPE, 0, value,
+		    (size_t)info.size) != (ssize_t)info.size)
+			@throw [OFGetItemAttributesFailedException
+			    exceptionWithIRI: IRI
+				       errNo: errno];
+
+		*data = [OFData dataWithItemsNoCopy: value
+					      count: (size_t)info.size
+				       freeWhenDone: true];
+		value = NULL;
+
+		if (type != NULL)
+			*type = [OFNumber numberWithUnsignedLong: info.type];
+	} @finally {
+		OFFreeMemory(value);
+		close(fd);
+	}
 # endif
+
+	[*data retain];
+	if (type != NULL)
+		[*type retain];
+
+	objc_autoreleasePoolPop(pool);
+
+	[*data autorelease];
+	if (type != NULL)
+		[*type autorelease];
+}
+
+- (void)setExtendedAttributeData: (OFData *)data
+			 andType: (id)type
+			 forName: (OFString *)name
+		     ofItemAtIRI: (OFIRI *)IRI
+{
+	void *pool = objc_autoreleasePoolPush();
+	OFString *path = IRI.fileSystemRepresentation;
+	OFStringEncoding encoding = [OFLocale encoding];
+	const char *cPath = [path cStringWithEncoding: encoding];
+	size_t size = data.count * data.itemSize;
+
+# if defined(OF_LINUX) || defined(OF_MACOS)
+	const char *cName = [name cStringWithEncoding: encoding];
+
+	if (type != nil)
+		@throw [OFNotImplementedException exceptionWithSelector: _cmd
+								 object: self];
+
+#  if defined(OF_LINUX)
+	if (lsetxattr(cPath, cName, data.items, size, 0) != 0) {
+#  elif defined(OF_MACOS)
+	if (setxattr(cPath, cName, data.items, size, 0, XATTR_NOFOLLOW) != 0) {
+#  endif
 		int errNo = errno;
 
 		/* TODO: Add an attribute (prefix?) for extended attributes? */
@@ -1650,6 +1882,76 @@ setExtendedAttributes(OFMutableFileAttributes attributes, OFIRI *IRI)
 		     failedAttribute: @""
 			       errNo: errNo];
 	}
+# elif defined(OF_FREEBSD) || defined(OF_NETBSD)
+	int namespace;
+	const char *cName;
+
+	if (size > SSIZE_MAX)
+		@throw [OFOutOfRangeException exception];
+
+	if (type != nil)
+		@throw [OFNotImplementedException exceptionWithSelector: _cmd
+								 object: self];
+
+	parseAttributeName(&name, &namespace);
+	cName = [name cStringWithEncoding: encoding];
+
+	if (extattr_set_link(cPath, namespace, cName, data.items, size) !=
+	    (ssize_t)size) {
+		int errNo = errno;
+
+		/* TODO: Add an attribute (prefix?) for extended attributes? */
+		@throw [OFSetItemAttributesFailedException
+		    exceptionWithIRI: IRI
+			  attributes: [OFDictionary dictionary]
+		     failedAttribute: @""
+			       errNo: errNo];
+	}
+# elif defined(OF_HAIKU)
+	const char *cName = [name cStringWithEncoding: encoding];
+	unsigned long long typeInt;
+	int fd;
+
+	if (type != nil && ![type isKindOfClass: [OFNumber class]])
+		@throw [OFInvalidArgumentException exception];
+
+	typeInt = (type != nil ? [type unsignedLongLongValue] : 0);
+	if (typeInt > UINT32_MAX)
+		@throw [OFInvalidArgumentException exception];
+
+	if (size > SSIZE_MAX)
+		@throw [OFOutOfRangeException exception];
+
+	if ((fd = open(cPath, O_WRONLY)) == -1) {
+		int errNo = errno;
+
+		/* TODO: Add an attribute (prefix?) for extended attributes? */
+		@throw [OFSetItemAttributesFailedException
+		    exceptionWithIRI: IRI
+			  attributes: [OFDictionary dictionary]
+		     failedAttribute: @""
+			       errNo: errNo];
+	}
+
+	@try {
+		if (fs_write_attr(fd, cName, (uint32_t)typeInt, 0,
+		    data.items, size) != (ssize_t)size) {
+			int errNo = errno;
+
+			/*
+			 * TODO: Add an attribute (prefix?) for extended
+			 *	 attributes?
+			 */
+			@throw [OFSetItemAttributesFailedException
+			    exceptionWithIRI: IRI
+				  attributes: [OFDictionary dictionary]
+			     failedAttribute: @""
+				       errNo: errNo];
+		}
+	} @finally {
+		close(fd);
+	}
+# endif
 
 	objc_autoreleasePoolPop(pool);
 }
@@ -1660,14 +1962,45 @@ setExtendedAttributes(OFMutableFileAttributes attributes, OFIRI *IRI)
 	void *pool = objc_autoreleasePoolPush();
 	OFString *path = IRI.fileSystemRepresentation;
 	OFStringEncoding encoding = [OFLocale encoding];
+	const char *cPath = [path cStringWithEncoding: encoding];
+# if defined(OF_LINUX) || defined(OF_MACOS)
+	const char *cName = [name cStringWithEncoding: encoding];
 
-# if defined(OF_LINUX)
-	if (lremovexattr([path cStringWithEncoding: encoding],
-	    [name cStringWithEncoding: encoding]) != 0) {
-# elif defined(OF_MACOS)
-	if (removexattr([path cStringWithEncoding: encoding],
-	    [name cStringWithEncoding: encoding], XATTR_NOFOLLOW) != 0) {
-# endif
+#  if defined(OF_LINUX)
+	if (lremovexattr(cPath, cName) != 0) {
+#  elif defined(OF_MACOS)
+	if (removexattr(cPath, cName, XATTR_NOFOLLOW) != 0) {
+#  endif
+		int errNo = errno;
+
+		/* TODO: Add an attribute (prefix?) for extended attributes? */
+		@throw [OFSetItemAttributesFailedException
+		    exceptionWithIRI: IRI
+			  attributes: [OFDictionary dictionary]
+		     failedAttribute: @""
+			       errNo: errNo];
+	}
+# elif defined(OF_FREEBSD) || defined(OF_NETBSD)
+	int namespace;
+	const char *cName;
+
+	parseAttributeName(&name, &namespace);
+	cName = [name cStringWithEncoding: encoding];
+
+	if (extattr_delete_link(cPath, namespace, cName) != 0) {
+		int errNo = errno;
+
+		/* TODO: Add an attribute (prefix?) for extended attributes? */
+		@throw [OFSetItemAttributesFailedException
+		    exceptionWithIRI: IRI
+			  attributes: [OFDictionary dictionary]
+		     failedAttribute: @""
+			       errNo: errNo];
+	}
+# elif defined(OF_HAIKU)
+	int fd;
+
+	if ((fd = open(cPath, O_WRONLY)) == -1) {
 		int errNo = errno;
 
 		/* TODO: Add an attribute (prefix?) for extended attributes? */
@@ -1678,6 +2011,25 @@ setExtendedAttributes(OFMutableFileAttributes attributes, OFIRI *IRI)
 			       errNo: errNo];
 	}
 
+	@try {
+		if (fs_remove_attr(fd, cName) != 0) {
+			int errNo = errno;
+
+			/*
+			 * TODO: Add an attribute (prefix?) for extended
+			 *	 attributes?
+			 */
+			@throw [OFSetItemAttributesFailedException
+			    exceptionWithIRI: IRI
+				  attributes: [OFDictionary dictionary]
+			     failedAttribute: @""
+				       errNo: errNo];
+		}
+	} @finally {
+		close(fd);
+	}
+# endif
+
 	objc_autoreleasePoolPop(pool);
 }
 #endif
