@@ -3,14 +3,18 @@
  *
  * All rights reserved.
  *
- * This file is part of ObjFW. It may be distributed under the terms of the
- * Q Public License 1.0, which can be found in the file LICENSE.QPL included in
- * the packaging of this file.
+ * This program is free software: you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License version 3.0 only,
+ * as published by the Free Software Foundation.
  *
- * Alternatively, it may be distributed under the terms of the GNU General
- * Public License, either version 2 or 3, which can be found in the file
- * LICENSE.GPLv2 or LICENSE.GPLv3 respectively included in the packaging of this
- * file.
+ * This program is distributed in the hope that it will be useful, but WITHOUT
+ * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+ * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
+ * version 3.0 for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public License
+ * version 3.0 along with this program. If not, see
+ * <https://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -1155,6 +1159,27 @@ merge(OFString *base, OFString *path)
 	return [ret autorelease];
 }
 
+- (OFString *)pathExtension
+{
+	void *pool = objc_autoreleasePoolPush();
+	OFString *ret, *fileName;
+	size_t pos;
+
+	fileName = self.lastPathComponent;
+	pos = [fileName rangeOfString: @"."
+			      options: OFStringSearchBackwards].location;
+	if (pos == OFNotFound || pos == 0) {
+		objc_autoreleasePoolPop(pool);
+		return @"";
+	}
+
+	ret = [fileName substringFromIndex: pos + 1];
+
+	[ret retain];
+	objc_autoreleasePoolPop(pool);
+	return [ret autorelease];
+}
+
 - (OFString *)query
 {
 	return _percentEncodedQuery.stringByRemovingPercentEncoding;
@@ -1313,6 +1338,30 @@ merge(OFString *base, OFString *path)
 	return IRI;
 }
 
+- (OFIRI *)IRIByDeletingLastPathComponent
+{
+	OFMutableIRI *IRI = [[self mutableCopy] autorelease];
+	[IRI deleteLastPathComponent];
+	[IRI makeImmutable];
+	return IRI;
+}
+
+- (OFIRI *)IRIByAppendingPathExtension: (OFString *)extension
+{
+	OFMutableIRI *IRI = [[self mutableCopy] autorelease];
+	[IRI appendPathExtension: extension];
+	[IRI makeImmutable];
+	return IRI;
+}
+
+- (OFIRI *)IRIByDeletingPathExtension
+{
+	OFMutableIRI *IRI = [[self mutableCopy] autorelease];
+	[IRI deletePathExtension];
+	[IRI makeImmutable];
+	return IRI;
+}
+
 - (OFIRI *)IRIByStandardizingPath
 {
 	OFMutableIRI *IRI = [[self mutableCopy] autorelease];
