@@ -3,14 +3,18 @@
  *
  * All rights reserved.
  *
- * This file is part of ObjFW. It may be distributed under the terms of the
- * Q Public License 1.0, which can be found in the file LICENSE.QPL included in
- * the packaging of this file.
+ * This program is free software: you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License version 3.0 only,
+ * as published by the Free Software Foundation.
  *
- * Alternatively, it may be distributed under the terms of the GNU General
- * Public License, either version 2 or 3, which can be found in the file
- * LICENSE.GPLv2 or LICENSE.GPLv3 respectively included in the packaging of this
- * file.
+ * This program is distributed in the hope that it will be useful, but WITHOUT
+ * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+ * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
+ * version 3.0 for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public License
+ * version 3.0 along with this program. If not, see
+ * <https://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -33,7 +37,6 @@
 #import "OFOutOfRangeException.h"
 #import "OFUnsupportedVersionException.h"
 
-@implementation OFLHAArchiveEntry
 static OFDate *
 parseMSDOSDate(uint32_t MSDOSDate)
 {
@@ -53,6 +56,7 @@ parseMSDOSDate(uint32_t MSDOSDate)
 					format: @"%Y-%m-%d %H:%M:%S"];
 }
 
+@implementation OFLHAArchiveEntry
 static void
 parseFileNameExtension(OFLHAArchiveEntry *entry, OFData *extension,
     OFStringEncoding encoding)
@@ -201,6 +205,22 @@ parseModificationDateExtension(OFLHAArchiveEntry *entry, OFData *extension,
 	    initWithTimeIntervalSince1970: modificationDate];
 }
 
+static void
+parseFileSizeExtension(OFLHAArchiveEntry *entry, OFData *extension,
+    OFStringEncoding encoding)
+{
+	uint64_t tmp;
+
+	if (extension.count != 17)
+		@throw [OFInvalidFormatException exception];
+
+	memcpy(&tmp, (char *)extension.items + 1, 8);
+	entry->_compressedSize = OFFromLittleEndian64(tmp);
+
+	memcpy(&tmp, (char *)extension.items + 9, 8);
+	entry->_uncompressedSize = OFFromLittleEndian64(tmp);
+}
+
 static bool
 parseExtension(OFLHAArchiveEntry *entry, OFData *extension,
     OFStringEncoding encoding, bool allowFileName)
@@ -219,6 +239,9 @@ parseExtension(OFLHAArchiveEntry *entry, OFData *extension,
 	case 0x3F:
 		function = parseCommentExtension;
 		break;
+	case 0x42:
+		function = parseFileSizeExtension;
+		break;
 	case 0x50:
 		function = parsePermissionsExtension;
 		break;
@@ -250,19 +273,25 @@ readExtensions(OFLHAArchiveEntry *entry, OFStream *stream,
 	size_t consumed = 0;
 
 	for (;;) {
-		uint16_t size;
+		uint32_t size;
 		OFData *extension;
 
-		size = [stream readLittleEndianInt16];
-		consumed += 2;
+		if (entry->_headerLevel == 3) {
+			size = [stream readLittleEndianInt32];
+			consumed += 4;
+		} else {
+			size = [stream readLittleEndianInt16];
+			consumed += 2;
+		}
 
 		if (size == 0)
 			break;
 
-		if (size < 2)
+		if (size < 2 || (entry->_headerLevel == 3 && size < 4))
 			@throw [OFInvalidFormatException exception];
 
-		extension = [stream readDataWithCount: size - 2];
+		extension = [stream readDataWithCount:
+		    size - (entry->_headerLevel == 3 ? 4 : 2)];
 		consumed += extension.count;
 
 		if (!parseExtension(entry, extension, encoding, allowFileName))
@@ -419,7 +448,8 @@ getFileNameAndDirectoryName(OFLHAArchiveEntry *entry, OFStringEncoding encoding,
 
 			objc_autoreleasePoolPop(pool);
 			break;
-		case 2:;
+		case 2:
+		case 3:;
 			uint32_t padding = 0;
 
 			_modificationDate = [[OFDate alloc]
@@ -428,11 +458,17 @@ getFileNameAndDirectoryName(OFLHAArchiveEntry *entry, OFStringEncoding encoding,
 			_CRC16 = [stream readLittleEndianInt16];
 			_operatingSystemIdentifier = [stream readInt8];
 
+			if (_headerLevel == 3)
+				/* Size of entire header */
+				padding = [stream readLittleEndianInt32];
+			else
+				padding = (header[1] << 8) | header[0];
+
 			/*
 			 * 21 for header, 2 for CRC16, 1 for operating system
 			 * identifier.
 			 */
-			padding = ((header[1] << 8) | header[0]) - 21 - 2 - 1;
+			padding -= 21 + 2 + 1;
 
 			padding -= readExtensions(self, stream, encoding, true);
 
@@ -625,6 +661,7 @@ getFileNameAndDirectoryName(OFLHAArchiveEntry *entry, OFStringEncoding encoding,
 	size_t fileNameLength, directoryNameLength;
 	uint16_t tmp16;
 	uint32_t tmp32;
+	uint64_t tmp64;
 	size_t headerSize;
 
 	if ([_compressionMethod cStringLengthWithEncoding:
@@ -636,7 +673,7 @@ getFileNameAndDirectoryName(OFLHAArchiveEntry *entry, OFStringEncoding encoding,
 
 	if (fileNameLength > UINT16_MAX - 3 ||
 	    directoryNameLength > UINT16_MAX - 3 ||
-	    _compressedSize > UINT32_MAX || _uncompressedSize > UINT32_MAX)
+	    _compressedSize > UINT64_MAX || _uncompressedSize > UINT64_MAX)
 		@throw [OFOutOfRangeException exception];
 
 	/* Length. Filled in after we're done. */
@@ -701,6 +738,20 @@ getFileNameAndDirectoryName(OFLHAArchiveEntry *entry, OFStringEncoding encoding,
 			 count: fileCommentLength];
 	}
 
+	/*
+	 * Always include the file size extension, as the header can be written
+	 * with size 0 initially and then rewritten with the actual size in
+	 * case the data to be archived is being streamed - but for that we
+	 * need to make sure we always have the space.
+	 */
+	tmp16 = OFToLittleEndian16(19);
+	[data addItems: &tmp16 count: sizeof(tmp16)];
+	[data addItem: "\x42"];
+	tmp64 = OFToLittleEndian64(_compressedSize);
+	[data addItems: &tmp64 count: sizeof(tmp64)];
+	tmp64 = OFToLittleEndian64(_uncompressedSize);
+	[data addItems: &tmp64 count: sizeof(tmp64)];
+
 	if (_POSIXPermissions != nil) {
 		tmp16 = OFToLittleEndian16(5);
 		[data addItems: &tmp16 count: sizeof(tmp16)];
