@@ -3,18 +3,20 @@
  *
  * All rights reserved.
  *
- * This file is part of ObjFW. It may be distributed under the terms of the
- * Q Public License 1.0, which can be found in the file LICENSE.QPL included in
- * the packaging of this file.
+ * This program is free software: you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License version 3.0 only,
+ * as published by the Free Software Foundation.
  *
- * Alternatively, it may be distributed under the terms of the GNU General
- * Public License, either version 2 or 3, which can be found in the file
- * LICENSE.GPLv2 or LICENSE.GPLv3 respectively included in the packaging of this
- * file.
+ * This program is distributed in the hope that it will be useful, but WITHOUT
+ * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+ * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
+ * version 3.0 for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public License
+ * version 3.0 along with this program. If not, see
+ * <https://www.gnu.org/licenses/>.
  */
 
-#define __NO_EXT_QNX
-
 #include "config.h"
 
 #include <errno.h>
@@ -96,11 +98,6 @@
 	[super dealloc];
 }
 
-- (bool)lowlevelIsAtEndOfStream
-{
-	OF_UNRECOGNIZED_SELECTOR
-}
-
 - (size_t)lowlevelReadIntoBuffer: (void *)buffer length: (size_t)length
 {
 	OF_UNRECOGNIZED_SELECTOR
@@ -111,6 +108,16 @@
 	OF_UNRECOGNIZED_SELECTOR
 }
 
+- (bool)lowlevelIsAtEndOfStream
+{
+	OF_UNRECOGNIZED_SELECTOR
+}
+
+- (bool)lowlevelHasDataInReadBuffer
+{
+	return false;
+}
+
 - (id)copy
 {
 	return [self retain];
@@ -701,7 +708,7 @@
 				ret = [OFString
 				    stringWithCString: _readBuffer
 					     encoding: encoding
-					      length: i + 1 - delimiterLength];
+					       length: i + 1 - delimiterLength];
 
 				_readBuffer += i + 1;
 				_readBufferLength -= i + 1;
@@ -1186,7 +1193,7 @@
 
 - (bool)hasDataInReadBuffer
 {
-	return (_readBufferLength > 0);
+	return (_readBufferLength > 0 || [self lowlevelHasDataInReadBuffer]);
 }
 
 - (bool)canBlock
