@@ -3,14 +3,18 @@
  *
  * All rights reserved.
  *
- * This file is part of ObjFW. It may be distributed under the terms of the
- * Q Public License 1.0, which can be found in the file LICENSE.QPL included in
- * the packaging of this file.
+ * This program is free software: you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License version 3.0 only,
+ * as published by the Free Software Foundation.
  *
- * Alternatively, it may be distributed under the terms of the GNU General
- * Public License, either version 2 or 3, which can be found in the file
- * LICENSE.GPLv2 or LICENSE.GPLv3 respectively included in the packaging of this
- * file.
+ * This program is distributed in the hope that it will be useful, but WITHOUT
+ * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+ * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
+ * version 3.0 for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public License
+ * version 3.0 along with this program. If not, see
+ * <https://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -998,7 +1002,7 @@ OF_SINGLETON_METHODS
 	OFIRI *IRI;
 
 	@try {
-		IRI = [OFIRI fileIRIWithPath: path];
+		IRI = [OFIRI fileIRIWithPath: path isDirectory: false];
 	} @catch (id e) {
 		[self release];
 		@throw e;
@@ -2340,14 +2344,24 @@ OF_SINGLETON_METHODS
 #if defined(HAVE_STRTOF_L) || defined(HAVE_USELOCALE)
 	const char *UTF8String = self.UTF8String;
 #else
-	/*
-	 * If we have no strtof_l, we have no other choice but to replace "."
-	 * with the locale's decimal point.
-	 */
 	OFString *decimalSeparator = [OFLocale decimalSeparator];
-	const char *UTF8String = [self
-	    stringByReplacingOccurrencesOfString: @"."
-				      withString: decimalSeparator].UTF8String;
+	const char *UTF8String;
+
+	if ([decimalSeparator isEqual: @"."])
+		UTF8String = self.UTF8String;
+	else
+		/*
+		 * If we have no strtof_l, we have no other choice than to
+		 * replace the locale's decimal point with something that will
+		 * be rejected and replacing "." with the locale's decimal
+		 * point.
+		 */
+		UTF8String = [[self
+		    stringByReplacingOccurrencesOfString: decimalSeparator
+					      withString: @"!"]
+		    stringByReplacingOccurrencesOfString: @"."
+					      withString: decimalSeparator]
+		    .UTF8String;
 #endif
 	char *endPtr = NULL;
 	float value;
@@ -2397,14 +2411,24 @@ OF_SINGLETON_METHODS
 #if defined(HAVE_STRTOD_L) || defined(HAVE_USELOCALE)
 	const char *UTF8String = self.UTF8String;
 #else
-	/*
-	 * If we have no strtod_l, we have no other choice but to replace "."
-	 * with the locale's decimal point.
-	 */
 	OFString *decimalSeparator = [OFLocale decimalSeparator];
-	const char *UTF8String = [self
-	    stringByReplacingOccurrencesOfString: @"."
-				      withString: decimalSeparator].UTF8String;
+	const char *UTF8String;
+
+	if ([decimalSeparator isEqual: @"."])
+		UTF8String = self.UTF8String;
+	else
+		/*
+		 * If we have no strtod_l, we have no other choice than to
+		 * replace the locale's decimal point with something that will
+		 * be rejected and replacing "." with the locale's decimal
+		 * point.
+		 */
+		UTF8String = [[self
+		    stringByReplacingOccurrencesOfString: decimalSeparator
+					      withString: @"!"]
+		    stringByReplacingOccurrencesOfString: @"."
+					      withString: decimalSeparator]
+		    .UTF8String;
 #endif
 	char *endPtr = NULL;
 	double value;
