@@ -3,14 +3,18 @@
  *
  * All rights reserved.
  *
- * This file is part of ObjFW. It may be distributed under the terms of the
- * Q Public License 1.0, which can be found in the file LICENSE.QPL included in
- * the packaging of this file.
+ * This program is free software: you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License version 3.0 only,
+ * as published by the Free Software Foundation.
  *
- * Alternatively, it may be distributed under the terms of the GNU General
- * Public License, either version 2 or 3, which can be found in the file
- * LICENSE.GPLv2 or LICENSE.GPLv3 respectively included in the packaging of this
- * file.
+ * This program is distributed in the hope that it will be useful, but WITHOUT
+ * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+ * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
+ * version 3.0 for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public License
+ * version 3.0 along with this program. If not, see
+ * <https://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -2041,9 +2045,9 @@ static const OFUnichar caseFoldingPage31[0x100] = {
 	8112, 8113, 8048, 8049, 8115, 0, 953, 0,
 	0, 0, 0, 0, 0, 0, 0, 0,
 	8050, 8051, 8052, 8053, 8131, 0, 0, 0,
-	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 912, 0, 0, 0, 0,
 	8144, 8145, 8054, 8055, 0, 0, 0, 0,
-	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 944, 0, 0, 0, 0,
 	8160, 8161, 8058, 8059, 8165, 0, 0, 0,
 	0, 0, 0, 0, 0, 0, 0, 0,
 	8056, 8057, 8060, 8061, 8179, 0, 0, 0,
@@ -2084,6 +2088,41 @@ static const OFUnichar caseFoldingPage171[0x100] = {
 	0, 0, 0, 0, 0, 0, 0, 0,
 };
 
+static const OFUnichar caseFoldingPage251[0x100] = {
+	0, 0, 0, 0, 0, 64262, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0,
+};
+
 const OFUnichar *const OFUnicodeUppercaseTable[0x1EA] = {
 	uppercasePage0, uppercasePage1, uppercasePage2, uppercasePage3,
 	uppercasePage4, uppercasePage5, emptyPage, emptyPage,
@@ -2546,7 +2585,7 @@ const OFUnichar *const OFUnicodeCaseFoldingTable[0x1EA] = {
 	emptyPage, emptyPage, emptyPage,
 	emptyPage, emptyPage, emptyPage,
 	emptyPage, emptyPage, emptyPage,
-	emptyPage, emptyPage, emptyPage,
+	emptyPage, emptyPage, caseFoldingPage251,
 	emptyPage, emptyPage, emptyPage,
 	lowercasePage255, emptyPage, emptyPage,
 	emptyPage, emptyPage, lowercasePage260,
